import numpy as np
import numpy.typing as npt
from nuplan.planning.simulation.trajectory.trajectory_sampling import TrajectorySampling
from navsim.common.dataclasses import PDMResults, Trajectory
from navsim.agents.expansion.scoring.pdm_scorer_expanded import (
    MultiMetricIndex,
    WeightedMetricIndex,
)

from navsim.agents.expansion.scoring.pdm_scorer_expanded import PDMScorerExpanded
from navsim.common.dataclasses import Trajectory
from navsim.evaluate.pdm_score import transform_trajectory, get_trajectory_as_array
from navsim.planning.metric_caching.metric_cache import MetricCache
from navsim.planning.simulation.planner.pdm_planner.simulation.pdm_simulator import (
    PDMSimulator,
)


def pdm_score_expanded(
        metric_cache: MetricCache,
        vocab_trajectory: npt.NDArray,
        pre_model_trajectory,
        future_sampling: TrajectorySampling,
        simulator: PDMSimulator,
        scorer: PDMScorerExpanded,
        expansion_only=True
) -> npt.NDArray:
    """
    Runs PDM-Score and saves results in dataclass.
    :param metric_cache: Metric cache dataclass
    :param vocab_trajectory: Predicted trajectory in ego frame.
    :return: Dataclass of PDM-Subscores.
    """

    initial_ego_state = metric_cache.ego_state
    pre_ego_state = metric_cache.others['pre_ego_status']

    transformed_ones = [transform_trajectory(Trajectory(pose, TrajectorySampling(
        time_horizon=4, interval_length=0.1
    )), initial_ego_state) for pose in vocab_trajectory]
    if pre_model_trajectory is not None:
        pre_pred_trajectory = transform_trajectory(pre_model_trajectory, pre_ego_state)
        pre_pre_states = get_trajectory_as_array(pre_pred_trajectory, future_sampling, pre_ego_state.time_point)[None]
        pre_simulated_states = simulator.simulate_proposals(pre_pre_states, pre_ego_state)
    else:
        pre_simulated_states = None
    pdm_states = get_trajectory_as_array(
        metric_cache.trajectory,
        future_sampling,
        initial_ego_state.time_point
    )[None]

    # pdm, vocab-0, vocab-1, ..., vocab-n
    all_states = [pdm_states]
    all_states += [
        get_trajectory_as_array(
            transformed,
            future_sampling,
            initial_ego_state.time_point
        )[None] for transformed in transformed_ones
    ]
    all_states = np.concatenate(all_states, axis=0)


    simulated_states = simulator.simulate_proposals(all_states, initial_ego_state)



    scores, scores_expand = scorer.score_proposals(
        simulated_states,
        pre_simulated_states,
        metric_cache.observation,
        metric_cache.centerline,
        metric_cache.route_lane_ids,
        metric_cache.drivable_area_map,
        metric_cache,
        expansion_only
    )
    if expansion_only:
        return {
            # expanded metrics
            # 'mAP': scorer.navigation_mAP,
            'lk': scorer._weighted_metrics[WeightedMetricIndex.LANE_KEEPING].astype(np.float16)[1:],
            # 'tl': scorer._multi_metrics[MultiMetricIndex.TRAFFIC_LIGHTS].astype(np.bool)[1:],
            # 'dr': scorer._multi_metrics[MultiMetricIndex.DRIVING_DIRECTION].astype(np.float16)[1:],
            # 'consistency': scorer._multi_metrics[WeightedMetricIndex.CONSISTENCY].astype(np.float16)[1:],
        }
    return {
        # ori metrics
        'noc': scorer._multi_metrics[MultiMetricIndex.NO_COLLISION].astype(np.float16)[1:],
        'da': scorer._multi_metrics[MultiMetricIndex.DRIVABLE_AREA].astype(np.bool)[1:],
        'dd': scorer._multi_metrics[MultiMetricIndex.DRIVING_DIRECTION].astype(np.float16)[1:],
        'ttc': scorer._weighted_metrics[WeightedMetricIndex.TTC].astype(np.bool)[1:],
        'progress': scorer._weighted_metrics[WeightedMetricIndex.PROGRESS].astype(np.float16)[1:],
        'comfort': scorer._weighted_metrics[WeightedMetricIndex.COMFORTABLE].astype(np.float16)[1:],
        # expanded metrics
        'mAP': scorer.navigation_mAP,
        'lk': scorer._weighted_metrics[WeightedMetricIndex.LANE_KEEPING].astype(np.float16)[1:],
        'tl': scorer._multi_metrics[MultiMetricIndex.TRAFFIC_LIGHTS].astype(np.bool)[1:],
        'dr': scorer._multi_metrics[MultiMetricIndex.DRIVING_DIRECTION].astype(np.float16)[1:],
        'consistency': scorer._weighted_metrics[WeightedMetricIndex.CONSISTENCY].astype(np.float16)[1:],
        'total': scores.astype(np.float16)[1:],
        'total_expand': scores_expand.astype(np.float16)[1:],
    }

def pdm_score_wcon(
        metric_cache: MetricCache,
        model_trajectory: Trajectory,
        pre_model_trajectory: Trajectory,
        future_sampling: TrajectorySampling,
        simulator: PDMSimulator,
        scorer: PDMScorerExpanded,
        expansion_only=True
) -> npt.NDArray:
    """
    Runs PDM-Score and saves results in dataclass.
    :param metric_cache: Metric cache dataclass
    :param model_trajectory: Predicted trajectory in ego frame.
    :return: Dataclass of PDM-Subscores.
    """

    initial_ego_state = metric_cache.ego_state
    pre_ego_state = metric_cache.others['pre_ego_status']

    pdm_trajectory = metric_cache.trajectory
    pred_trajectory = transform_trajectory(model_trajectory, initial_ego_state)
    pre_pred_trajectory = transform_trajectory(pre_model_trajectory, pre_ego_state)

    pdm_states, pred_states = (
        get_trajectory_as_array(pdm_trajectory, future_sampling, initial_ego_state.time_point),
        get_trajectory_as_array(pred_trajectory, future_sampling, initial_ego_state.time_point),
    )
    pre_pre_states = get_trajectory_as_array(pre_pred_trajectory, future_sampling, pre_ego_state.time_point)[None]

    trajectory_states = np.concatenate([pdm_states[None, ...], pred_states[None, ...]], axis=0)

    simulated_states = simulator.simulate_proposals(trajectory_states, initial_ego_state)
    pre_simulated_states = simulator.simulate_proposals(pre_pre_states, pre_ego_state)

    scores, scores_expand = scorer.score_proposals(
        simulated_states,
        pre_simulated_states,
        metric_cache.observation,
        metric_cache.centerline,
        metric_cache.route_lane_ids,
        metric_cache.drivable_area_map,
        metric_cache,
        expansion_only,
    )

    # TODO: Refactor & add / modify existing metrics.
    pred_idx = 0

    no_at_fault_collisions = scorer._multi_metrics[MultiMetricIndex.NO_COLLISION, pred_idx]
    drivable_area_compliance = scorer._multi_metrics[MultiMetricIndex.DRIVABLE_AREA, pred_idx]
    driving_direction_compliance = scorer._multi_metrics[
        MultiMetricIndex.DRIVING_DIRECTION, pred_idx
    ]

    ego_progress = scorer._weighted_metrics[WeightedMetricIndex.PROGRESS, pred_idx]
    time_to_collision_within_bound = scorer._weighted_metrics[WeightedMetricIndex.TTC, pred_idx]
    comfort = scorer._weighted_metrics[WeightedMetricIndex.COMFORTABLE, pred_idx]
    lane_keeping = scorer._weighted_metrics[WeightedMetricIndex.LANE_KEEPING, pred_idx]
    traffic_light = scorer._multi_metrics[MultiMetricIndex.TRAFFIC_LIGHTS, pred_idx]
    consistency = scorer._weighted_metrics[WeightedMetricIndex.CONSISTENCY, pred_idx]

    # score = scores[pred_idx]
    if expansion_only:
        return {
            # expanded metrics
            'mAP': 1.0,
            'lk': lane_keeping,
            'tl': traffic_light,
            'dr': driving_direction_compliance,
            'consistency': consistency,
        }
    return {
        # ori metrics
        'noc': no_at_fault_collisions,
        'da': drivable_area_compliance,
        'dd': driving_direction_compliance,
        'ttc': time_to_collision_within_bound,
        'progress': ego_progress,
        'comfort': comfort,
        # expanded metrics
        'mAP': 1.0,
        'lk': lane_keeping,
        'tl': traffic_light,
        'dr': driving_direction_compliance,
        'consistency': consistency,
        'total': scores[pred_idx],
        'total_expand': scores_expand[pred_idx],
    }

def pdm_score(
        metric_cache: MetricCache,
        model_trajectory: Trajectory,
        future_sampling: TrajectorySampling,
        simulator: PDMSimulator,
        scorer: PDMScorerExpanded,
        expansion_only=True
) -> npt.NDArray:
    """
    Runs PDM-Score and saves results in dataclass.
    :param metric_cache: Metric cache dataclass
    :param model_trajectory: Predicted trajectory in ego frame.
    :return: Dataclass of PDM-Subscores.
    """

    initial_ego_state = metric_cache.ego_state

    pdm_trajectory = metric_cache.trajectory
    pred_trajectory = transform_trajectory(model_trajectory, initial_ego_state)

    pdm_states, pred_states = (
        get_trajectory_as_array(pdm_trajectory, future_sampling, initial_ego_state.time_point),
        get_trajectory_as_array(pred_trajectory, future_sampling, initial_ego_state.time_point),
    )

    trajectory_states = np.concatenate([pdm_states[None, ...], pred_states[None, ...]], axis=0)

    simulated_states = simulator.simulate_proposals(trajectory_states, initial_ego_state)

    scores, scores_expand = scorer.score_proposals(
        simulated_states,
        None,
        metric_cache.observation,
        metric_cache.centerline,
        metric_cache.route_lane_ids,
        metric_cache.drivable_area_map,
        metric_cache,
        expansion_only,
    )

    # TODO: Refactor & add / modify existing metrics.
    pred_idx = 1

    no_at_fault_collisions = scorer._multi_metrics[MultiMetricIndex.NO_COLLISION, pred_idx]
    drivable_area_compliance = scorer._multi_metrics[MultiMetricIndex.DRIVABLE_AREA, pred_idx]
    driving_direction_compliance = scorer._multi_metrics[
        MultiMetricIndex.DRIVING_DIRECTION, pred_idx
    ]

    ego_progress = scorer._weighted_metrics[WeightedMetricIndex.PROGRESS, pred_idx]
    time_to_collision_within_bound = scorer._weighted_metrics[WeightedMetricIndex.TTC, pred_idx]
    comfort = scorer._weighted_metrics[WeightedMetricIndex.COMFORTABLE, pred_idx]
    lane_keeping = scorer._weighted_metrics[WeightedMetricIndex.LANE_KEEPING, pred_idx]
    traffic_light = scorer._multi_metrics[MultiMetricIndex.TRAFFIC_LIGHTS, pred_idx]

    # score = scores[pred_idx]
    if expansion_only:
        return {
            # expanded metrics
            'mAP': 1.0,
            'lk': lane_keeping,
            'tl': traffic_light,
            'dr': driving_direction_compliance,
        }
    return {
        # ori metrics[图片]
        'noc': no_at_fault_collisions,
        'da': drivable_area_compliance,
        'dd': driving_direction_compliance,
        'ttc': time_to_collision_within_bound,
        'progress': ego_progress,
        'comfort': comfort,
        # expanded metrics
        'mAP': 1.0,
        'lk': lane_keeping,
        'tl': traffic_light,
        'dr': driving_direction_compliance,
        'total': scores[pred_idx],
        'total_expand': scores_expand[pred_idx],
    }